// Common code between platforms
// Copyright  Glenn Fiedler
// http://www.pixeltoaster.com

#include <ctime>

namespace PixelToaster
{
	// derive your platform's display implementation from this and it will handle all the mundane details for you

	class DisplayAdapter : public DisplayInterface
	{
	public:

		DisplayAdapter()
		{
			_listener = 0;
			defaults();
		}

		~DisplayAdapter()
		{
			close();
			_listener = 0;
		}

		void listener(Listener *listener)
		{
			_listener = listener;
		}

		Listener* listener() const
		{
			return _listener;
		}

        bool open(const char title[], int width, int height, Mode mode, Output output)
		{
			close();

			strncpy(_title, title, sizeof(_title));
			_width = width;
			_height = height;
			_mode = mode;
			_output = output;			// note: you need to interpret default output yourself, and set it to either "windowed" or "fullscreen"
			
			return true;
		}

        void close()
		{
			defaults();
		}

		bool update(integer32 pixels[])
		{
			if (pixels)
				return update(pixels, 0);
			else
				return false;
		}

		bool update(Pixel pixels[])
		{
			if (pixels)
				return update(0, pixels);
			else
				return false;
		}

		const char* title() const
		{
			return _title;
		}

		int width() const
		{
			return _width;
		}

		int height() const
		{
			return _height;
		}

		Mode mode() const
		{
			return _mode;
		}

		Output output() const
		{
			return _output;
		}

	protected:

		// note: override this "unified" update to implement your display update.
		// only one of the pointers will be non-null, this allows you to avoid
		// duplicating update code between integer and floating point update methods.

		virtual bool update(integer32 *integerPixels, Pixel *floatingPointPixels) = 0;

		// this defaults is virtual, override it to add your own defaults
		// but make sure you always call the superclass defaults in your overridden function!
		// note: due to c++ constructor oddities, make sure you also call defaults in your own 
		// display ctor, otherwise it will only call this base version because the vtable is not 
		// setup yet. see the windows display for an example of this.

		virtual void defaults()
		{
			_title[0] = 0;
			_width = 0;
			_height = 0;
			_mode = Mode::FloatingPointColor;
			_output = Output::Default;
		}

		// switch to windowed output.
		// override this to implement your own switch.

		virtual bool windowed()
		{
			_output = Output::Windowed;
			return true;
		}

		// switch to fullscreen output.
		// override this to implement your own switch.

		virtual bool fullscreen()
		{
			_output = Output::Fullscreen;
			return true;
		}

	private:

		char _title[1024];
		int _width;
		int _height;
		Mode _mode;
		Output _output;
		Listener *_listener;
	};

	// portable timer implementation for platforms without a specific high res timer

	class PortableTimer : public PixelToaster::TimerInterface
	{
	public:
		
		PortableTimer()
		{
			_resolution = 1.0 / CLOCKS_PER_SEC;
			reset();
		}
		
		void reset()
		{
			_time = 0;
			_timeCounter = clock();
			_deltaCounter = _timeCounter;
		}
		
		double time()
		{
			clock_t counter = std::clock();
			double delta = (counter - _timeCounter) * _resolution;
			_timeCounter = counter;
			_time += delta;
			return _time;
		}
		
		double delta()
		{
			clock_t counter = std::clock();
			double delta = (counter - _deltaCounter) * _resolution;
			_deltaCounter = counter;
			return delta;
		}
		
		double resolution()
		{
			return _resolution;
		}
		
		void wait(double seconds)
		{
			clock_t start = std::clock();
			clock_t finish = start + clock_t(seconds / _resolution);
			while (std::clock()<finish);
		}
		
	private:
		
		double _time;               ///< current time in seconds
		double _resolution;			///< timer resolution in seconds
		clock_t _timeCounter;		///< time counter in clocks
		clock_t _deltaCounter;		///< delta counter in clocks
	};
}
