// Conversion routines
// Copyright  Glenn Fiedler
// http://www.pixeltoaster.com

#ifndef PIXEL_TOASTER_CONVERSION
#define PIXEL_TOASTER_CONVERSION

#include <memory.h>
#include <float.h>
#include <stdio.h>

namespace PixelToaster
{
    inline integer32 clampedfraction8(float value)
    {
        int &integer = (int&)value;

        if (integer<=0)
            return 0;
        else if (integer>=0x3F800000)
            return 0x07F8000;

        value += 1.0f;

        return integer & 0x07F8000;
    }

    inline void convertPixelToXRGB8888(const Pixel source[], integer32 destination[], unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            const integer32 r = clampedfraction8(source[i].r) << 1;
            const integer32 g = clampedfraction8(source[i].g) >> 7;
            const integer32 b = clampedfraction8(source[i].b) >> 15;

            destination[i] = r | g | b;
        }
    }

    inline void convertXRGB8888ToPixel(const integer32 source[], Pixel destination[], unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            destination[i].r = float((source[i]>>16)&0xFF)*1.0f/255.0f;
            destination[i].g = float((source[i]>>8)&0xFF)*1.0f/255.0f;
            destination[i].b = float((source[i])&0xFF)*1.0f/255.0f;
            destination[i].a = 0.0f;
        }
    }

    inline void convertPixelToXBGR8888(const Pixel source[], integer32 destination[], unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            const integer32 r = clampedfraction8(source[i].r) >> 15;
            const integer32 g = clampedfraction8(source[i].g) >> 7;
            const integer32 b = clampedfraction8(source[i].b) << 1;

            destination[i] = r | g | b;
        }
    }

    inline void convertXBGR8888ToPixel(const integer32 source[], Pixel destination[], unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            destination[i].r = float((source[i])&0xFF)*1.0f/255.0f;
            destination[i].g = float((source[i]>>8)&0xFF)*1.0f/255.0f;
            destination[i].b = float((source[i]>>16)&0xFF)*1.0f/255.0f;
            destination[i].a = 0.0f;
        }
    }

    void convertPixelToRGB888(const Pixel source[], integer8 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            const integer32 r = clampedfraction8(source[i].r) >> 15;
            const integer32 g = clampedfraction8(source[i].g) >> 15;
            const integer32 b = clampedfraction8(source[i].b) >> 15;

            destination[0] = r;
            destination[1] = g;
            destination[2] = b;

            destination += 3;
        }
    }

    void convertRGB888ToPixel(const integer8 *source, Pixel destination[], unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 r = source[0];
            integer32 g = source[1];
            integer32 b = source[2];

            destination[i].r = float(r) * 1.0f/255.0f;
            destination[i].g = float(g) * 1.0f/255.0f;
            destination[i].b = float(b) * 1.0f/255.0f;
            destination[i].a = 0;

            source += 3;
        }
    }

    void convertPixelToBGR888(const Pixel source[], integer8 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            const integer32 r = clampedfraction8(source[i].r) >> 15;
            const integer32 g = clampedfraction8(source[i].g) >> 15;
            const integer32 b = clampedfraction8(source[i].b) >> 15;

            destination[0] = b;
            destination[1] = g;
            destination[2] = r;

            destination += 3;
        }
    }

    void convertBGR888ToPixel(const integer8 *source, Pixel destination[], unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 b = source[0];
            integer32 g = source[1];
            integer32 r = source[2];

            destination[i].r = float(r) * 1.0f/255.0f;
            destination[i].g = float(g) * 1.0f/255.0f;
            destination[i].b = float(b) * 1.0f/255.0f;
            destination[i].a = 0;

            source += 3;
        }
    }

    inline integer32 clampedfraction6(float value)
    {
        int &integer = (int&)value;

        if (integer<=0)
            return 0;
        else if (integer>=0x3F800000)
            return 0x07E0000;

        value += 1.0f;

        return integer & 0x07E0000;
    }

    inline integer32 clampedfraction5(float value)
    {
        int &integer = (int&)value;

        if (integer<=0)
            return 0;
        else if (integer>=0x3F800000)
            return 0x07C0000;

        value += 1.0f;

        return integer & 0x07C0000;
    }

    void convertPixelToRGB565(const Pixel source[], integer16 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            const integer32 r = clampedfraction5(source[i].r) >> 7;
            const integer32 g = clampedfraction6(source[i].g) >> 12;
            const integer32 b = clampedfraction5(source[i].b) >> 18;

            destination[i] = (integer16) (r | g | b);
        }
    }

    void convertRGB565ToPixel(const integer16 *source, Pixel destination[], unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 color = (integer32) source[i];

            integer32 r = (color & 0x0000F800) >> 11;
            integer32 g = (color & 0x000007E0) >> 5;
            integer32 b = color & 0x0000001F;

            destination[i].r = float(r) * 1.0f / 31.0f;
            destination[i].g = float(g) * 1.0f / 63.0f;
            destination[i].b = float(b) * 1.0f / 31.0f;
            destination[i].a = 0;
        }
    }

    void convertPixelToBGR565(const Pixel source[], integer16 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            const integer32 r = clampedfraction5(source[i].r) >> 18;
            const integer32 g = clampedfraction6(source[i].g) >> 12;
            const integer32 b = clampedfraction5(source[i].b) >> 7;

            destination[i] = (integer16) (r | g | b);
        }
    }

    void convertBGR565ToPixel(const integer16 *source, Pixel destination[], unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 color = (integer32) source[i];

            integer32 b = (color & 0x0000F800) >> 11;
            integer32 g = (color & 0x000007E0) >> 5;
            integer32 r = color & 0x0000001F;

            destination[i].r = float(r) * 1.0f / 31.0f;
            destination[i].g = float(g) * 1.0f / 63.0f;
            destination[i].b = float(b) * 1.0f / 31.0f;
            destination[i].a = 0;
        }
    }

    void convertPixelToXRGB1555(const Pixel source[], integer16 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            const integer32 r = clampedfraction5(source[i].r) >> 8;
            const integer32 g = clampedfraction5(source[i].g) >> 13;
            const integer32 b = clampedfraction5(source[i].b) >> 18;

            destination[i] = (integer16) (r | g | b);
        }
    }

    void convertXRGB1555ToPixel(const integer16 *source, Pixel destination[], unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 color = (integer32) source[i];

            integer32 r = (color & 0x00007C00) >> 10;
            integer32 g = (color & 0x000003E0) >> 5;
            integer32 b = color & 0x0000001F;

            destination[i].r = float(r) * 1.0f / 31.0f;
            destination[i].g = float(g) * 1.0f / 31.0f;
            destination[i].b = float(b) * 1.0f / 31.0f;
            destination[i].a = 0;
        }
    }

    void convertPixelToXBGR1555(const Pixel source[], integer16 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            const integer32 r = clampedfraction5(source[i].r) >> 18;
            const integer32 g = clampedfraction5(source[i].g) >> 13;
            const integer32 b = clampedfraction5(source[i].b) >> 8;

            destination[i] = (integer16) (r | g | b);
        }
    }

    void convertXBGR1555ToPixel(const integer16 *source, Pixel destination[], unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 color = (integer32) source[i];

            integer32 b = (color & 0x00007C00) >> 10;
            integer32 g = (color & 0x000003E0) >> 5;
            integer32 r = color & 0x0000001F;

            destination[i].r = float(r) * 1.0f / 31.0f;
            destination[i].g = float(g) * 1.0f / 31.0f;
            destination[i].b = float(b) * 1.0f / 31.0f;
            destination[i].a = 0;
        }
    }

    // integer to integer converters

    void convertXRGB8888ToXBGR8888(const integer32 *source, integer32 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 r = (source[i] & 0x00FF0000) >> 16;
            integer32 g = (source[i] & 0x0000FF00);
            integer32 b = (source[i] & 0x000000FF) << 16;
            destination[i] = r | g | b;
        }
    }

    void convertXBGR8888ToXRGB8888(const integer32 *source, integer32 *destination, unsigned int count)
    {
        convertXRGB8888ToXBGR8888(source, destination, count);
    }

    void convertXRGB8888ToRGB888(const integer32 *source, integer8 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            destination[0] = (integer8) ( (source[i] & 0x00FF0000) >> 16 );
            destination[1] = (integer8) ( (source[i] & 0x0000FF00) >> 8 );
            destination[2] = (integer8) ( (source[i] & 0x000000FF) );
            destination += 3;
        }
    }

    void convertRGB888ToXRGB8888(const integer8 *source, integer32 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 r = source[0];
            integer32 g = source[1];
            integer32 b = source[2];

            destination[i] = (r<<16) | (g<<8) | b;

            source += 3;
        }
    }

    void convertXRGB8888ToBGR888(const integer32 *source, integer8 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            destination[0] = (integer8) ( (source[i] & 0x000000FF) );
            destination[1] = (integer8) ( (source[i] & 0x0000FF00) >> 8 );
            destination[2] = (integer8) ( (source[i] & 0x00FF0000) >> 16 );
            destination += 3;
        }
    }

    void convertBGR888ToXRGB8888(const integer8 *source, integer32 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 r = source[0];
            integer32 g = source[1];
            integer32 b = source[2];

            destination[i] = (b<<16) | (g<<8) | r;

            source += 3;
        }
    }

    void convertXRGB8888ToRGB565(const integer32 *source, integer16 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 r = (source[i] & 0x00F80000) >> 8;
            integer32 g = (source[i] & 0x0000FC00) >> 5;
            integer32 b = (source[i] & 0x000000F8) >> 3;
            
            destination[i] = (integer16) (r | g | b);
        }
    }

    void convertRGB565ToXRGB8888(const integer16 *source, integer32 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 color = (integer32) source[i];
            
            integer32 r = (color & 0x0000F800) << 8;
            integer32 g = (color & 0x000007E0) << 5;
            integer32 b = (color & 0x0000001F) << 3;

            destination[i] = r | g | b;
        }
    }

    void convertXRGB8888ToBGR565(const integer32 *source, integer16 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 r = (source[i] & 0x00F80000) >> 19;
            integer32 g = (source[i] & 0x0000FC00) >> 5;
            integer32 b = (source[i] & 0x000000F8) << 8;     // right shift seems slow!

            destination[i] = (integer16) (r | g | b);
        }
    }

    void convertBGR565ToXRGB8888(const integer16 *source, integer32 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 color = (integer32) source[i];

            integer32 r = ((color & 0x0000F800) << 8) >> 16;
            integer32 g = (color & 0x000007E0) << 5;
            integer32 b = ((color & 0x0000001F)) << 3 << 16;

            destination[i] = r | g | b;
        }
    }

    void convertXRGB8888ToXRGB1555(const integer32 *source, integer16 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 r = (source[i] & 0x00F80000) >> 9;
            integer32 g = (source[i] & 0x0000F800) >> 6;
            integer32 b = (source[i] & 0x000000F8) >> 3;

            destination[i] = (integer16) (r | g | b);
        }
    }

    void convertXRGB1555ToXRGB8888(const integer16 *source, integer32 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 color = (integer32) source[i];

            integer32 r = (color & 0x00007C00) << 9;
            integer32 g = (color & 0x000003E0) << 6;
            integer32 b = (color & 0x0000001F) << 3;

            destination[i] = r | g | b;
        }
    }

    void convertXRGB8888ToXBGR1555(const integer32 *source, integer16 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 r = (source[i] & 0x00F80000) >> 19;
            integer32 g = (source[i] & 0x0000F800) >> 6;
            integer32 b = (source[i] & 0x000000F8) << 7;

            destination[i] = (integer16) (r | g | b);
        }
    }

    void convertXBGR1555ToXRGB8888(const integer16 *source, integer32 *destination, unsigned int count)
    {
        for (unsigned int i=0; i<count; i++)
        {
            integer32 color = (integer32) source[i];

            integer32 r = (color & 0x00007C00) >> 7;
            integer32 g = (color & 0x000003E0) << 6;
            integer32 b = (color & 0x0000001F) << 19;

            destination[i] = r | g | b;
        }
    }

    // copy converters

    void convertXRGB8888ToXRGB8888(const integer32 *source, integer32 *destination, unsigned int count)
    {
        memcpy(destination, source, count*4);
    }

    void convertPixelToPixel(const Pixel source[], Pixel destination[], unsigned int count)
    {
        memcpy(destination, source, count*16);
    }

    // converter classes

    class ConverterAdapter : public Converter
    {
        void begin() {};
        void end() {}
    };

    class MMXConverter : public ConverterAdapter
    {
        void begin()
        {
            // emms
        }

        void end()
        {
            // normal floating point operations
        }
    };

    class Converter_PixelToPixel : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertPixelToPixel((const Pixel*)source, (Pixel*)destination, pixels);
        }
    };

    class Converter_PixelToXRGB8888 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertPixelToXRGB8888((const Pixel*)source, (integer32*)destination, pixels);
        }
    };

    class Converter_PixelToXBGR8888 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertPixelToXBGR8888((const Pixel*)source, (integer32*)destination, pixels);
        }
    };

    class Converter_PixelToRGB888 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertPixelToRGB888((const Pixel*)source, (integer8*)destination, pixels);
        }
    };

    class Converter_PixelToBGR888 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertPixelToBGR888((const Pixel*)source, (integer8*)destination, pixels);
        }
    };

    class Converter_PixelToRGB565 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertPixelToRGB565((const Pixel*)source, (integer16*)destination, pixels);
        }
    };

    class Converter_PixelToBGR565 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertPixelToBGR565((const Pixel*)source, (integer16*)destination, pixels);
        }
    };

    class Converter_PixelToXRGB1555 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertPixelToXRGB1555((const Pixel*)source, (integer16*)destination, pixels);
        }
    };

    class Converter_PixelToXBGR1555 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertPixelToXBGR1555((const Pixel*)source, (integer16*)destination, pixels);
        }
    };

    class Converter_XRGB8888ToPixel : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertXRGB8888ToPixel((const integer32*)source, (Pixel*)destination, pixels);
        }
    };

    class Converter_XRGB8888ToXRGB8888 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertXRGB8888ToXRGB8888((const integer32*)source, (integer32*)destination, pixels);
        }
    };

    class Converter_XRGB8888ToXBGR8888 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertXRGB8888ToXBGR8888((const integer32*)source, (integer32*)destination, pixels);
        }
    };

    class Converter_XRGB8888ToRGB888 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertXRGB8888ToRGB888((const integer32*)source, (integer8*)destination, pixels);
        }
    };

    class Converter_XRGB8888ToBGR888 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertXRGB8888ToBGR888((const integer32*)source, (integer8*)destination, pixels);
        }
    };

    class Converter_XRGB8888ToRGB565 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertXRGB8888ToRGB565((const integer32*)source, (integer16*)destination, pixels);
        }
    };

    class Converter_XRGB8888ToBGR565 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertXRGB8888ToBGR565((const integer32*)source, (integer16*)destination, pixels);
        }
    };

    class Converter_XRGB8888ToXRGB1555 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertXRGB8888ToXRGB1555((const integer32*)source, (integer16*)destination, pixels);
        }
    };

    class Converter_XRGB8888ToXBGR1555 : public ConverterAdapter
    {
        void convert(const void *source, void *destination, int pixels)
        {
            convertXRGB8888ToXBGR1555((const integer32*)source, (integer16*)destination, pixels);
        }
    };
}

#endif