// Pixel Toaster interface
// Copyright  Glenn Fiedler
// http://www.pixeltoaster.com

#ifndef PIXEL_TOASTER_INTERFACE
#define PIXEL_TOASTER_INTERFACE

#if 0//def _WIN32
    #ifdef PIXEL_TOASTER_DLL
    #define PIXEL_TOASTER_API __declspec(dllexport)
    #else
    #define PIXEL_TOASTER_API __declspec(dllimport)
    #endif
#else
    #define PIXEL_TOASTER_API
#endif

namespace PixelToaster
{
    /// %Mode class.
    /// This is just a simple wrapper class around an enumeration.
    /// It is used to select between truecolor and floating point color when opening a display.
    /// See Display::open for details.

    class Mode
    {
    public:

        /// %Mode enumeration.

        enum Enumeration
        {
            TrueColor,                  ///< truecolor mode. pixels are represented by 32 bit integers.
            FloatingPointColor,         ///< floating point color mode. pixels are represented by floating point values. See the Pixel object for details.
        };

        /// Default constructor.
        /// Sets the enumeration value to TrueColor.

        Mode()
        {
            enumeration = TrueColor;
        }

        /// %Mode constructor
        /// @param enumeration the enumeration value.

        Mode(Enumeration enumeration)
        {
            this->enumeration = enumeration;
        }

        /// Cast from mode object to enumeration.
        /// Allows you to treat this class as if it was the enumeration itself.
        /// This enables the ==, != operators, and the use of mode objects in a switch statement.

        operator Enumeration() const
        {
            return enumeration;
        }

    private:

        Enumeration enumeration;
    };

    /// %Format class.
    /// This is just a simple wrapper class around an enumeration.
    /// It is used to describe various memory layouts for storing pixels in memory.
    /// You do not normally need to use this directly, as %Pixel Toaster will 
    /// automatically convert from XRGB8888 (truecolor), or Pixels (floating point color)
    /// to the best pixel format available on the display.

    class Format
    {
    public:

        /// %Format enumeration.

        enum Enumeration
        {
            Unknown,        ///< unknown pixel format.
            XRGB8888,       ///< 32 bit truecolor. this is the native pixel format in Mode::TrueColor.
            XBGR8888,       ///< 32 bit truecolor in BGR order.
            RGB888,         ///< 24 bit truecolor.
            BGR888,         ///< 24 bit truecolor in BGR order.
            RGB565,         ///< 16 bit hicolor.
            BGR565,         ///< 16 bit hicolor in BGR order.
            XRGB1555,       ///< 15 bit hicolor.
            XBGR1555,       ///< 15 bit hicolor in BGR order.
            Pixels,         ///< floating point color. this is the native pixel format in Mode::FloatingPointColor.
        };

        /// Default constructor.
        /// Sets the enumeration value to Unknown.

        Format()
        {
            enumeration = Unknown;
        }

        /// %Format constructor
        /// @param enumeration the enumeration value.

        Format(Enumeration enumeration)
        {
            this->enumeration = enumeration;
        }

        /// Cast from format object to enumeration.
        /// Allows you to treat this class as if it was the enumeration itself.
        /// This enables the ==, != operators, and the use of format objects in a switch statement.

        operator Enumeration() const
        {
            return enumeration;
        }

    private:

        Enumeration enumeration;
    };

    /// %Output class.
    /// This is just a simple wrapper class around an enumeration.
    /// It is used to select between fullscreen and windowed display output.
    /// See Display::open for details.

    class Output
    {
    public:

        /// %Output enumeration.

        enum Enumeration
        {
            Default,            ///< default output. let the display choose the best output method.
            Windowed,           ///< windowed output. output pixels to a window.
            Fullscreen,         ///< fullscreen output. switch to a fullscreen display mode.
        };

        /// Default constructor.
        /// Sets the enumeration value to Default.

        Output()
        {
            enumeration = Default;
        }

        /// %Output constructor
        /// @param enumeration the enumeration value.

        Output(Enumeration enumeration)
        {
            this->enumeration = enumeration;
        }

        /// Cast from output object to enumeration.
        /// Allows you to treat this class as if it was the enumeration itself.
        /// This enables the ==, != operators, and the use of output objects in a switch statement.

        operator Enumeration() const
        {
            return enumeration;
        }

    private:

        Enumeration enumeration;
    };

    /// %Mouse class.
    /// Represents the current mouse state such as the current cursor position,
    /// and what mouse buttons are being pressed.

    class Mouse
    {
    public:

        /// %Mouse button state.
        /// Stores information about multiple mouse buttons being pressed.

        class Buttons
        {
        public:
            bool left;        ///< true if left button is pressed.
            bool middle;      ///< true if middle button is pressed.
            bool right;       ///< true if right button is pressed.
        };

        Buttons buttons;      ///< mouse button state
        int x;                ///< current mouse cursor x position. ranges from 0 to display width - 1, from left to right.
        int y;                ///< current mouse cursor y position. ranges from 0 to display height - 1, from top to bottom.
    };

    /// %Key class.
    /// This is just a simple wrapper class around an enumeration of all key codes.
    /// For all intents and purposes you may treat an instance of this class as if
    /// it was the key code enumeration value itself.

    class Key
    {
    public:

        /// Key code.

        enum Code
        {
            Enter          = '\n',      ///< enter key
            BackSpace      = '\b',      ///< backspace key
            Tab            = '\t',      ///< tab key
            Cancel         = 0x03,      ///< cancel key
            Clear          = 0x0C,      ///< clear key
            Shift          = 0x10,      ///< shift key
            Control        = 0x11,      ///< control key
            Alt            = 0x12,      ///< alt key
            Pause          = 0x13,      ///< pause key
            CapsLock       = 0x14,      ///< capslock key
            Escape         = 0x1B,      ///< escape key
            Space          = 0x20,      ///< space key
            PageUp         = 0x21,      ///< page up key
            PageDown       = 0x22,      ///< page down key
            End            = 0x23,      ///< end key
            Home           = 0x24,      ///< home key
            Left           = 0x25,      ///< left key
            Up             = 0x26,      ///< up arrow key
            Right          = 0x27,      ///< right arrow key
            Down           = 0x28,      ///< down arrow key
            Comma          = 0x2C,      ///< comma key ','
            Period         = 0x2E,      ///< period key '.'
            Slash          = 0x2F,      ///< slash key '/'
            Zero           = 0x30,      ///< zero key
            One            = 0x31,      ///< one key
            Two            = 0x32,      ///< two key
            Three          = 0x33,      ///< three key
            Four           = 0x34,      ///< four key
            Five           = 0x35,      ///< five key
            Six            = 0x36,      ///< six key
            Seven          = 0x37,      ///< seven key
            Eight          = 0x38,      ///< eight key
            Nine           = 0x39,      ///< nine key
            SemiColon      = 0x3B,      ///< semicolon key ';'
            Equals         = 0x3D,      ///< equals key '='
            A              = 0x41,      ///< a key
            B              = 0x42,      ///< b key
            C              = 0x43,      ///< c key
            D              = 0x44,      ///< d key
            E              = 0x45,      ///< e key
            F              = 0x46,      ///< f key
            G              = 0x47,      ///< g key
            H              = 0x48,      ///< h key
            I              = 0x49,      ///< i key
            J              = 0x4A,      ///< j key
            K              = 0x4B,      ///< k key
            L              = 0x4C,      ///< l key
            M              = 0x4D,      ///< m key
            N              = 0x4E,      ///< n key
            O              = 0x4F,      ///< o key
            P              = 0x50,      ///< p key
            Q              = 0x51,      ///< q key
            R              = 0x52,      ///< r key
            S              = 0x53,      ///< s key
            T              = 0x54,      ///< t key
            U              = 0x55,      ///< u key
            V              = 0x56,      ///< v key
            W              = 0x57,      ///< w key
            X              = 0x58,      ///< x key
            Y              = 0x59,      ///< y key
            Z              = 0x5A,      ///< z key
            OpenBracket    = 0x5B,      ///< open bracket key '['
            BackSlash      = 0x5C,      ///< back slash key '\'
            CloseBracket   = 0x5D,      ///< close bracket key ']'
            NumPad0        = 0x60,      ///< numpad 0 key
            NumPad1        = 0x61,      ///< numpad 1 key
            NumPad2        = 0x62,      ///< numpad 2 key
            NumPad3        = 0x63,      ///< numpad 3 key
            NumPad4        = 0x64,      ///< numpad 4 key
            NumPad5        = 0x65,      ///< numpad 5 key
            NumPad6        = 0x66,      ///< numpad 6 key
            NumPad7        = 0x67,      ///< numpad 7 key
            NumPad8        = 0x68,      ///< numpad 8 key
            NumPad9        = 0x69,      ///< numpad 9 key
            Multiply       = 0x6A,      ///< multiply key '*'
            Add            = 0x6B,      ///< add key '+'
            Separator      = 0x6C,      ///< separator key '-'
            Subtract       = 0x6D,      ///< subtract key '-'
            Decimal        = 0x6E,      ///< decimal key '.'
            Divide         = 0x6F,      ///< divide key '/'
            F1             = 0x70,      ///< F1 key
            F2             = 0x71,      ///< F2 key
            F3             = 0x72,      ///< F3 key
            F4             = 0x73,      ///< F4 key
            F5             = 0x74,      ///< F5 key
            F6             = 0x75,      ///< F6 key
            F7             = 0x76,      ///< F7 key
            F8             = 0x77,      ///< F8 key
            F9             = 0x78,      ///< F9 key
            F10            = 0x79,      ///< F10 key
            F11            = 0x7A,      ///< F11 key
            F12            = 0x7B,      ///< F12 key
            Delete         = 0x7F,      ///< delete key
            NumLock        = 0x90,      ///< numlock key
            ScrollLock     = 0x91,      ///< scroll lock key
            PrintScreen    = 0x9A,      ///< print screen key
            Insert         = 0x9B,      ///< insert key
            Help           = 0x9C,      ///< help key
            Meta           = 0x9D,      ///< meta key
            BackQuote      = 0xC0,      ///< backquote key
            Quote          = 0xDE,      ///< quote key
            Final          = 0x18,      ///< final key
            Convert        = 0x1C,      ///< convert key
            NonConvert     = 0x1D,      ///< non convert key
            Accept         = 0x1E,      ///< accept key
            ModeChange     = 0x1F,      ///< mode change key
            Kana           = 0x15,      ///< kana key
            Kanji          = 0x19,      ///< kanji key
            Undefined      = 0x0        ///< undefined key
        };

        /// Default constructor.
        /// Sets key code to undefined.

        Key()
        {
            code = Undefined;
        }

        /// Key constructor.
        /// @param code the key code.

        Key(Code code)
        {
            this->code = code;
        }

        /// Cast from key object to code.
        /// Allows you to treat this class as if it was the code itself.
        /// This enables the ==, != operators, and the use of key objects in a switch statement.

        operator Code() const
        {
            return code;
        }

    private:

        Code code;
    };

    typedef unsigned int integer32;
    typedef unsigned short integer16;
    typedef unsigned char integer8;

    PIXEL_TOASTER_API class DisplayInterface* createDisplay();
    PIXEL_TOASTER_API class TimerInterface* createTimer();
    PIXEL_TOASTER_API class Converter* requestConverter(Format source, Format destination);

    class DisplayInterface
    {
    public:

        virtual ~DisplayInterface() {};

        virtual bool open(const char title[], int width, int height, Mode mode = Mode::TrueColor, Output output = Output::Default) = 0;
        virtual void close() = 0;

        virtual void listener(class Listener *listener) = 0;
		virtual class Listener* listener() const = 0;

        virtual bool update(class Pixel pixels[]) = 0;
        virtual bool update(integer32 pixels[]) = 0;

        virtual const char* title() const = 0;
        virtual int width() const = 0;
        virtual int height() const = 0;
        virtual Mode mode() const = 0;
        virtual Output output() const = 0;
    };

    /// %Display class.
    /// This is the main class in %Pixel Toaster.
    /// The display class creates the output window or enters the display mode and copy pixels to the screen.
    /// Its also responsible for generating input events and calling back via the Listener interface.

    class Display
    {
    public:

        /// Default constructor.
        /// Creates the display object but does not open a display window or fullscreen mode.
        /// Make sure you call Display::open before updating pixels with Display::update.

        Display()
        {
            internal = createDisplay();
        }

        /// Destructor.
        /// Closes the display if it is still open.

        ~Display()
        {
            if (internal)
            {
                delete internal;
                internal = 0;
            }
        }

        /// Open display.
        /// Opens a display window or fullscreen mode. 
        /// You must call this method successfully before you can update pixels to the display.
        /// @param title the title of the display window or fullscreen application.
        /// @param width the width of the display in pixels.
        /// @param height the height of the display in pixels.
        /// @param mode the mode of operation for the display. you can choose between true color mode and floating point color mode.
        /// @param output the output type of the display. you can choose between windowed output and fullscreen output, or you can leave it up to the display by passing in default.
        /// @returns true if the display open was successful.

        bool open(const char title[], int width, int height, Mode mode = Mode::TrueColor, Output output = Output::Default)
        {
            if (internal)
                return internal->open(title, width, height, mode, output);
            else
                return false;
        }

        /// Close display.
        /// Closes the display window or leaves fullscreen mode.
        /// This method is safe to call even if the display not open.

        void close()
        {
            if (internal)
                internal->close();
        }

        /// Register a listener object.
        /// Implement the Listener interface and pass in an pointer to an instance of your object to 
		/// recieve display events such as keyboard and mouse input.
        /// @param listener the listener object. pass in 0 if you want to remove the current listener.

        void listener(class Listener *listener)
        {
            if (internal)
                internal->listener(listener);
        }

		/// Get current listener object.
		/// @returns the currrent listener object pointer. null if there is no listener.

		class Listener* listener() const
		{
			if (internal)
				return internal->listener();
			else
				return 0;
		}

        /// Update display with floating point pixels.
        /// The input pixels must be a linear array of size width x height, where width and height are the
        /// dimensions of the display as per the most recent successful call to Display::open.
        /// You can calculate the offset for a pixel in the linear array as follows: int offset = width*y + x;
        /// The pixels in the array are in Format::Pixels, for details see the Pixel class.
        /// Note that this is the natural update method to call when the display was opened in Mode::FloatingPointColor,
        /// however it is safe to call this method even when operating in Mode::TrueColor if you wish.
        /// @param pixels the pixels to copy to the screen.
        /// @returns true if the update was successful.

        bool update(class Pixel pixels[])
        {
            if (internal)
                return internal->update(pixels);
            else
                return false;
        }

        /// Update display with truecolor pixels.
        /// The input pixels must be a linear array of size width x height, where width and height are the
        /// dimensions of the display as per the most recent successful call to Display::open.
        /// You can calculate the offset for a pixel in the linear array as follows: int offset = width*y + x;
        /// The pixels in the array are in Format::ARGB8888. You can pack integer r,g,b values in range [0,255]
        /// into a color integer32 as follows: integer32 color = (r<<16) | (g<<8) | b;
        /// Note that this is the natural update method to call when the display was opened in Mode::TrueColor,
        /// however it is safe to call this method even when operating in Mode::FloatingPointColor if you wish.
        /// @param pixels the pixels to copy to the screen.
        /// @returns true if the update was successful.

        bool update(integer32 pixels[])
        {
            if (internal)
                return internal->update(pixels);
            else
                return false;
        }

        /// Get display title

        const char* title() const
        {
            if (internal)
                return internal->title();
            else
                return "";
        }

        /// Get display width

        int width() const
        {
            if (internal)
                return internal->width();
            else
                return 0;
        }

        /// Get display height

        int height() const
        {
            if (internal)
                return internal->height();
            else
                return 0;
        }

        /// Get display mode

        Mode mode() const
        {
            if (internal)
                return internal->mode();
            else
                return Mode::TrueColor;
        }

        /// Get display output.
        
        Output output() const
        {
            if (internal)
                return internal->output();
            else
                return Output::Default;
        }

    private:

        DisplayInterface *internal;
    };

    class TimerInterface
    {
    public:
        virtual ~TimerInterface() {};
        virtual void reset() = 0;
        virtual double time() = 0;
        virtual double delta() = 0;
        virtual double resolution() = 0;
        virtual void wait(double seconds) = 0;
    };

    /// %Timer class.
    /// A high resolution timer used for measuring elapsed time 
    /// and calculating delta time per update.
    /// Uses doubles instead of floats to ensure maximum precision is available.

    class Timer
    {
    public:

        /// Default constructor.
        /// Sets current time to 0.0

        Timer()
        {
            internal = createTimer();
        }

        /// Destructor.

        ~Timer()
        {
            delete internal;
        }

        /// Reset time.             
        /// Current time is set to 0.0

        void reset()
        {
            if (internal)
                internal->reset();
        }

        /// Get current time.
        /// @returns current time in seconds.

        double time()
        {
            if (internal)
                return internal->time();
            else
                return 0.0;
        }

        /// Get delta time elapsed.
        /// @returns delta time since last call to Timer::delta or Timer::reset.

        double delta()
        {
            if (internal)
                return internal->delta();
            else
                return 0.0;
        }

        /// Get timer resolution.
        /// The timer resolution is the smallest change in time that can be detected.
        /// @returns timer resolution in seconds.

        double resolution()
        {
            if (internal)
                return internal->resolution();
            else
                return 0.0;
        }

        /// Wait for a period of time.
        /// @param seconds the number of seconds to wait before returning from this method.

        void wait(double seconds)
        {
            if (internal)
                internal->wait(seconds);
        }

    private:

        TimerInterface *internal;
    };

    /// %Listener interface.
    /// You must implement this interface and register an instance
    /// with Display::listener in order to recieve events such as 
    /// keyboard and mouse input.

    class Listener
    {
    public:

        /// On key down.
        /// Called once only when a key is pressed and held.
        /// @param key the key event data.

        virtual void onKeyDown(Key key) = 0;

        /// On key pressed.
        /// Called multiple times while a key is pressed and held including the initial event.
        /// @param key the key event data.

        virtual void onKeyPressed(Key key) = 0;

        /// On key up.
        /// Called when a key is released.
        /// @param key the key event data.

        virtual void onKeyUp(Key key) = 0;

        /// On mouse button down.
        /// Called once only when a mouse button is initially pressed.
        /// @param mouse the mouse event data.

        virtual void onMouseButtonDown(Mouse mouse) = 0;

        /// On mouse button up.
        /// Called when a mouse button is released.
        /// @param mouse the mouse event data.

        virtual void onMouseButtonUp(Mouse mouse) = 0;

        /// On mouse move.
        /// Called when the mouse is moved.
        /// @param mouse the mouse event data.

        virtual void onMouseMove(Mouse mouse) = 0;

        /// On activate.
        /// Called when the window is activated or deactivated.
        /// @param active true if the window is being activated, false if it is being deactivated.

        virtual void onActivate(bool active) = 0;

        /// On close.
        /// Called when the window has been requested to close by the user.
        /// You are responsible for responding to this event by quitting
        /// your application, otherwise the window will stay open and the
        /// application will keep running.

        virtual void onClose() = 0;
    };

    /// %Listener adapter.
    /// This is a helper class that you can derive from if you
    /// do not wish to implement every method in the Listener
    /// interface manually. Each method in the interface is implemented 
    /// as a stub function, so you are free to only override the methods
    /// corresponding to events that you want to recieve.

    class ListenerAdapter : public Listener
    {
        void onKeyDown(Key key) {}
        void onKeyPressed(Key key) {}
        void onKeyUp(Key key) {}
        void onMouseButtonUp(Mouse mouse) {}
        void onMouseButtonDown(Mouse mouse) {}
        void onMouseMove(Mouse mouse) {}
        void onActivate(bool active) {}
        void onClose() {}
    };

    /// %Pixel class.
    /// This is the native representation of pixels when in Mode::FloatingPointColor.
    /// It corresponds to the pixel format Format::Pixels.
    /// White is represented by all setting all color channels to 1.0f,
    /// Black when all values are 0.0f. Values outside the range [0,1] are allowed, 
    /// and are automatically clamped when displaying the pixels. 
    /// This allows you to exploit the high dynamic range provided by floating point color. 
    /// Also included is an alpha channel per pixel which you are free to use however you like.
    /// For example, it could be as a z-buffer for a software renderer, or simply as 4 bytes 
    /// of storage per pixel for any purpose. 
    /// If you dont have any use for the alpha value, its safe to just ignore it. 
        
    class Pixel
    {
    public:
        float a;        ///< alpha channel
        float b;        ///< blue channel
        float g;        ///< green channel
        float r;        ///< red channel
    };          

    class Converter
    {
    public:
        virtual void begin() = 0;
        virtual void convert(const void *source, void *destination, int pixels) = 0;
        virtual void end() = 0;
    };
}

#ifdef _WIN32
#define __WINDOWS__
#endif

#endif
